{-# LANGUAGE FlexibleContexts #-}
{-# LANGUAGE GADTs #-}

--
-- Multi variable polynomials
--

module Polynomial (
	Variable, Monomial,                           -- reexports from Monomial
	Polynomial(coeff), RationalFunction,          -- data types
	poly, cnst, fromMono, var,                    -- constructors
	x, y, z,                                      -- predefined variables
	appl,                                         -- application
	assertions,                                   -- type assertions
	getVars, getVarsR, onevar, novar, occursIn,   -- analysis methods
	psort, clift, mlift, d,                       -- operations
	lc, lm, lt, deg, total_deg, normalize, renameVars,
	subst, substR,
	division,
) where

import Control.Exception
import Data.Function
import Data.List
import Data.Ratio

import Monomial


------------------------------------
-- Data type
------------------------------------

-- | A polynomial is a list of monomials and their coefficients
data Polynomial t where
	Poly :: (Eq t, Num t) => { coeff :: [(t, Monomial)] } -> Polynomial t

-- | A rational function is represented by its numerator and denominator
--   polynomial
type RationalFunction t = Ratio (Polynomial t)

-- | Sort the monomials by their degree
psort :: (Eq t, Num t, Ord Monomial) => Polynomial t -> Polynomial t
psort = Poly . reverse . sortBy (compare `on` snd) . coeff

-- | A safe constructor for Polynomials (guarantees the assertions)
poly :: (Eq t, Num t, Ord Monomial) => [(t, Monomial)] -> Polynomial t
poly = psort. sum . map (\(a, m) -> cnst a * fromMono m) . filter ((/= 0) . fst)

-- | The constant polynomial
cnst :: (Eq t, Num t) => t -> Polynomial t
cnst 0 = Poly []
cnst a = Poly [(a, mono [])]

-- | Create a polynomial consisting of a single monomial
fromMono :: (Eq t, Num t) => Monomial -> Polynomial t
fromMono m = Poly [(1, m)]

-- | General variable constructor
var :: (Eq t, Num t) => Variable -> Polynomial t
var v = Poly [(1, varm v)]

-- | First few variable instances
x, y, z :: (Eq t, Num t) => Polynomial t
x = var "x"
y = var "y"
z = var "z"

-- | Convert the polynomial into a function
appl :: Num t => (Variable -> t) -> Polynomial t -> t
appl f = sum . map (\(a, m) -> a * applm f m) . coeff


------------------------------------
-- Assertions
------------------------------------

-- | Assertions for Polynomials
ass1, ass2, ass3, ass4, ass5 :: (Eq t, Num t) => Polynomial t -> Bool
ass6 :: (Eq t, Num t, Ord Monomial) => Polynomial t -> Bool
ass1 (Poly ms) = all (monoass1 . snd) ms -- first monomial assertion
ass2 (Poly ms) = all (monoass2 . snd) ms -- second monomial assertion
ass3 (Poly ms) = all (monoass3 . snd) ms -- third monomial assertion
ass4 (Poly ms) = all ((/= 0) . fst) ms   -- non-zero coefficients
ass5 (Poly ms) = nubBy ((==) `on` snd) ms == ms -- every monomial occurs only once
ass6 p         = psort p == p     -- monomials are ordered

-- | All the assertions combined
assertions :: (Eq t, Num t, Ord Monomial) => Polynomial t -> Bool
assertions p = all ($ p) [ass1, ass2, ass3, ass4, ass5, ass6]


------------------------------------
-- Class instances
------------------------------------

{-
instance Show a => Show (Ratio a) where
	show r | denominator r == 1 = numerator r
	       | otherwise          = numerator r ++ " % " ++ denominator r
-}

instance (Eq t, Num t) => Eq (Polynomial t) where
	(Poly ms) == (Poly ms') = ms == ms' -- Polynomials are ordered

instance (Eq t, Num t, Show t) => Show (Polynomial t) where
	show (Poly []) = "0"
	show (Poly ms) = intercalate " + " $ map showTerm ms where
		showTerm (a,m) | null (unMono m) = show a
		               | a /= 1          = '(' : show a ++ ")*" ++ show m
		               | otherwise       = show m

instance (Eq t, Num t, Ord Monomial) => Num (Polynomial t) where
	negate p               = clift negate p
	abs p                  = clift abs p
	signum p               = clift signum p
	fromInteger i          = cnst $ fromInteger i
	(Poly ms) + (Poly ms') = psort $ Poly $
				[(a, m) | (a, m) <- ms, not $ m `elem` map snd ms', a /= 0] ++
				[(a, m) | (a, m) <- ms', not $ m `elem` map snd ms, a /= 0] ++
				[(a + b, m) | (a, m) <- ms, (b, m') <- ms', m == m', a + b /= 0]
	(Poly ms) * (Poly ms') = sum
				[Poly [(a * b, m `multm` m')] | (a, m) <- ms, (b, m') <- ms']

instance (Eq t, Num t, Ord t, Ord Monomial) => Ord (Polynomial t) where
	compare p q = if null l then EQ else head l where
		l = dropWhile (== EQ)
			$ map (uncurry compare)
			$ (zip `on` (map flipTuple . coeff)) p q
		flipTuple (a,b) = (b,a)

instance (Eq t, Fractional t, Ord Monomial) => Fractional (Polynomial t) where
	fromRational = cnst . fromRational
	p / q | r /= 0    = error "Polynomial.(/): divisor does not divide the divident."
	      | otherwise = s
		where ([s],r) = division p [q]

instance (Eq t, Real t, Ord Monomial) => Real (Polynomial t) where
	toRational p | novar p   = toRational $ fst $ head $ coeff p
	             | otherwise = undefined
	-- Needed for Integral

instance (Eq t, Num t, Ord Monomial) => Enum (Polynomial t) where
	toEnum   = undefined -- Needed for Integral
	fromEnum = undefined -- Needed for Integral

instance (Eq t, Fractional t, Real t, Ord Monomial) => Integral (Polynomial t) where
	divMod p q  = (s,r) where ([s],r) = division p [q]
	quotRem     = divMod -- ?!?
	toInteger   = undefined


------------------------------------
-- Analysis methods
------------------------------------

-- | Find the variable with highest index; -1 for constant polynomials
getVars :: Polynomial t -> [Variable]
getVars = nub . concat . map (map fst . unMono . snd) . coeff

getVarsR :: (Fractional t, Real t, Ord Monomial) => RationalFunction t -> [Variable]
getVarsR f = nub $ getVars (numerator f) ++ getVars (denominator f)

-- | Does the polynomial have just one variable?
onevar :: Polynomial t -> Bool
onevar = (== 1) . length . getVars

-- | Is the polynomial constant?
novar :: Polynomial t -> Bool
novar = null . getVars

-- | Does a variable occur in the polynomial?
occursIn :: Variable -> Polynomial t -> Bool
occursIn v = any (v `elem`) . map (map fst . unMono . snd) . coeff


------------------------------------
-- Operations
------------------------------------

-- | Lift a function on the polynomial's coefficients
clift :: (Eq u, Num u) => (t -> u) -> Polynomial t -> Polynomial u
clift f = Poly . filter ((/= 0) . fst) . map (\(a, m) -> (f a, m)) . coeff

-- | Lift a function to the polynomial's monomials
mlift :: (Eq t, Num t, Ord Monomial) => (Monomial -> Monomial) -> Polynomial t -> Polynomial t
mlift f = poly . map (\(a, m) -> (a, f m)) . coeff

-- | The leading coefficient
lc :: (Eq t, Num t, Ord Monomial) => Polynomial t -> t
lc (Poly []) = 0
lc (Poly ms) = fst $ head ms

-- | The leading monomial
lm :: (Eq t, Num t, Ord Monomial) => Polynomial t -> Monomial
lm (Poly []) = mono []
lm (Poly ms) = snd $ head ms

-- | The leading term as polynomial
lt :: (Eq t, Num t, Ord Monomial) => Polynomial t -> Polynomial t
lt p = if lc p == 0 then Poly [] else Poly [(lc p, lm p)]

-- | Degree of the polynomial
deg :: (Eq t, Num t, Ord Monomial) => Polynomial t -> [(Variable, Integer)]
deg = undefined -- TODO: how should it be defined?

-- | Compute the maximum total degree of a polynomial
total_deg :: Polynomial t -> Integer
total_deg = maximum . map (sum . map snd . unMono . snd) . coeff

-- | Multiply with a number such that the leading coefficent is 1
normalize :: (Eq t, Fractional t, Ord Monomial) => Polynomial t -> Polynomial t
normalize 0 = 0
normalize p = (cnst $ 1 / lc p) * p

-- | Rename the variables in a polynomial
renameVars :: (Eq t, Num t, Ord Monomial) => (Variable -> Variable) -> Polynomial t -> Polynomial t
renameVars f = mlift (mono . map (\(v,i) -> (f v,i)) . unMono)

-- | Substitute a polynomial for a variable
subst :: (Eq t, Fractional t, Ord Monomial) =>
         Variable -> Polynomial t -> Polynomial t -> Polynomial t
subst v p = sum . map (\(a,m) -> cnst a * subst' m) . coeff
  where
    subst' m = fromMono (mono $ filter ((/= v) . fst) $ unMono m) * p^(getExp m v)

-- | Subst for rational functions
substR :: (Eq t, Fractional t, Real t, Ord Monomial) =>
          Variable -> RationalFunction t -> RationalFunction t -> RationalFunction t
substR v f g = (sum $ map (\(a,m) -> cnst a % 1 * substR' m) $ coeff $ numerator g)
               / (sum $ map (\(a,m) -> cnst a % 1 * substR' m) $ coeff $ denominator g)
  where
    substR' m = fromMono (mono $ filter ((/= v) . fst) $ unMono m) % 1 * f^(getExp m v)

-- | Partial derivative
d :: (Eq t, Num t, Ord Monomial) => Variable -> Polynomial t -> Polynomial t
d v = poly . map deriv . coeff where
	deriv (a,m)
		| v `elem` map fst (unMono m) =
			(a * fromIntegral (getExp m v), m `divm` varm v)
		| otherwise = (0,m)


------------------------------------
-- Division algorithm
------------------------------------

-- | Division algorithm for multi-variable polynomials
division :: (Eq t, Fractional t, Ord Monomial) => Polynomial t -> [Polynomial t] -> ([Polynomial t], Polynomial t)
division p qs
	| any (== 0) qs = error "Polynomial.division: divisors must not be zero"
	| p == 0        = ([0 | q <- qs], 0)
	| null divisors = let (as, r) = division (droplt p) qs in (as, r + lt p)
						-- nothing divides the leading term
	| otherwise = let
			(i, q)  = head divisors -- pick a divisor
			q'      = Poly [(lc p, lm p `divm` lm q)] -- lt(q' * q) == lt p * lc q
			ok      = lt (q' * q) == lt p * cnst (lc q)
			(as, r) = division (clift (*lc q) (droplt p) - q' * droplt q) qs
					in assert ok (update i (+ clift (/lc q) q') as, r)
	where
		divisors = filter (\(_, q)-> lm q `divides` lm p) (zip [0..] qs)
			-- divisors of p in qs
		droplt = poly . tail . coeff -- drop the leading term

-- | Change an element in a list by index
update :: Int -> (a -> a) -> [a] -> [a]
update n _ []     = error "Polynomial.replace: index too large"
update 0 f (x:xs) = f x : xs
update n f (x:xs) = x : update (n - 1) f xs
-- should be more efficient than
-- take (n-1) xs ++ f (xs !! n) ++ drop n xs
